from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class SOLJSON(typing.TypedDict):
    kind: typing.Literal["SOL"]


class BTCJSON(typing.TypedDict):
    kind: typing.Literal["BTC"]


class ETHJSON(typing.TypedDict):
    kind: typing.Literal["ETH"]


class APTJSON(typing.TypedDict):
    kind: typing.Literal["APT"]


class ARBJSON(typing.TypedDict):
    kind: typing.Literal["ARB"]


class UNDEFINEDJSON(typing.TypedDict):
    kind: typing.Literal["UNDEFINED"]


@dataclass
class SOL:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "SOL"

    @classmethod
    def to_json(cls) -> SOLJSON:
        return SOLJSON(
            kind="SOL",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SOL": {},
        }


@dataclass
class BTC:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "BTC"

    @classmethod
    def to_json(cls) -> BTCJSON:
        return BTCJSON(
            kind="BTC",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BTC": {},
        }


@dataclass
class ETH:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "ETH"

    @classmethod
    def to_json(cls) -> ETHJSON:
        return ETHJSON(
            kind="ETH",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ETH": {},
        }


@dataclass
class APT:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "APT"

    @classmethod
    def to_json(cls) -> APTJSON:
        return APTJSON(
            kind="APT",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "APT": {},
        }


@dataclass
class ARB:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "ARB"

    @classmethod
    def to_json(cls) -> ARBJSON:
        return ARBJSON(
            kind="ARB",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ARB": {},
        }


@dataclass
class UNDEFINED:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "UNDEFINED"

    @classmethod
    def to_json(cls) -> UNDEFINEDJSON:
        return UNDEFINEDJSON(
            kind="UNDEFINED",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "UNDEFINED": {},
        }


AssetKind = typing.Union[SOL, BTC, ETH, APT, ARB, UNDEFINED]
AssetJSON = typing.Union[SOLJSON, BTCJSON, ETHJSON, APTJSON, ARBJSON, UNDEFINEDJSON]


def from_decoded(obj: dict) -> AssetKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "SOL" in obj:
        return SOL()
    if "BTC" in obj:
        return BTC()
    if "ETH" in obj:
        return ETH()
    if "APT" in obj:
        return APT()
    if "ARB" in obj:
        return ARB()
    if "UNDEFINED" in obj:
        return UNDEFINED()
    raise ValueError("Invalid enum object")


def from_json(obj: AssetJSON) -> AssetKind:
    if obj["kind"] == "SOL":
        return SOL()
    if obj["kind"] == "BTC":
        return BTC()
    if obj["kind"] == "ETH":
        return ETH()
    if obj["kind"] == "APT":
        return APT()
    if obj["kind"] == "ARB":
        return ARB()
    if obj["kind"] == "UNDEFINED":
        return UNDEFINED()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "SOL" / borsh.CStruct(),
    "BTC" / borsh.CStruct(),
    "ETH" / borsh.CStruct(),
    "APT" / borsh.CStruct(),
    "ARB" / borsh.CStruct(),
    "UNDEFINED" / borsh.CStruct(),
)



from __future__ import annotations

import logging
import os
from dataclasses import dataclass

from anchorpy import EventParser, Idl, Program, Provider, Wallet
from solana.rpc.async_api import AsyncClient
from solders.pubkey import Pubkey

from zetamarkets_py import constants, pda, utils
from zetamarkets_py.market import Market
from zetamarkets_py.types import Asset, Network
from zetamarkets_py.zeta_client.accounts.pricing import Pricing
from zetamarkets_py.zeta_client.accounts.state import State

idl_path = os.path.join(os.path.dirname(__file__), "idl/zeta.json")
with open(idl_path, "r") as f:
    idl = Idl.from_json(f.read())


@dataclass
class Exchange:
    # Initialize
    connection: AsyncClient
    program_id: Pubkey
    program: Program
    state: State
    pricing: Pricing
    markets: dict[Asset, Market]

    _event_parser: EventParser

    _state_address: Pubkey
    _pricing_address: Pubkey
    _serum_authority_address: Pubkey
    _mint_authority_address: Pubkey

    _logger: logging.Logger

    @classmethod
    async def load(
        cls,
        network: Network,
        connection: AsyncClient,
        assets: list[Asset] = Asset.all(),
        log_level: int = logging.CRITICAL,
    ) -> "Exchange":
        program_id = constants.ZETA_PID[network]
        provider = Provider(connection, Wallet.dummy())
        program = Program(idl, program_id, provider)
        _event_parser = EventParser(program_id, program.coder)

        # Accounts
        state_address = pda.get_state_address(program_id)
        state = await State.fetch(connection, state_address, connection.commitment, program_id=program_id)
        if state is None:
            raise Exception(f"State not found at {state_address}")

        pricing_address = pda.get_pricing_address(program_id)
        pricing = await Pricing.fetch(connection, pricing_address, connection.commitment, program_id=program_id)
        if pricing is None:
            raise Exception(f"Pricing not found at {pricing_address}")

        # Addresses
        _serum_authority_address = pda.get_serum_authority_address(program_id)
        _mint_authority_address = pda.get_mint_authority_address(program_id)

        markets = {
            asset: await Market.load(network, connection, asset, pricing.markets[asset.to_index()]) for asset in assets
        }

        # not currently used
        logger = utils.create_logger(f"{__name__}.{cls.__name__}", log_level)

        instance = cls(
            connection=connection,
            program_id=program_id,
            program=program,
            state=state,
            pricing=pricing,
            markets=markets,
            _event_parser=_event_parser,
            _state_address=state_address,
            _pricing_address=pricing_address,
            _serum_authority_address=_serum_authority_address,
            _mint_authority_address=_mint_authority_address,
            _logger=logger,
        )

        return instance

    @property
    def endpoint(self) -> str:
        return self.connection._provider.endpoint_uri

    @property
    def assets(self) -> list[Asset]:
        return list(self.markets.keys())

    # Priority Fees
    def toggle_auto_priority_fee(self):
        self._use_auto_priority_fee = not self._use_auto_priority_fee

    # TODO: add auto priority fee
    # TODO: add to solders
    # def update_auto_fee(self):
    #     account_list = []

    #     # Query the most written-to accounts
    #     # Note: getRecentPrioritizationFees() will account for global fees too if no one is writing to our accs
    #     for market in self.markets.values():
    #         account_list.append(market.address.perp_sync_queue_address)

    #     try:
    #         data = requests.post(
    #             self.endpoint,
    #             json={
    #                 "jsonrpc": "2.0",
    #                 "id": 1,
    #                 "method": "getRecentPrioritizationFees",
    #                 "params": [[account_list]],
    #             },
    #         )

    #         fees = sorted(
    #             [obj["prioritizationFee"] for obj in data.json()["result"]],
    #             key=lambda x: x["slot"],
    #             reverse=True,
    #         )[
    #             :20
    #         ]  # Grab the latest 20

    #         median = statistics.median(fees)
    #         self.priority_fee = min(median, self._auto_priority_fee_upper_limit)
    #         print(f"AutoUpdate priority fee. New fee = {self.priority_fee} microlamports per compute unit")
    #     except Exception as e:
    #         print(f"updateAutoFee failed {e}")



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeZetaReferralsRewardsWalletAccounts(typing.TypedDict):
    state: Pubkey
    referrals_rewards_wallet: Pubkey
    usdc_mint: Pubkey
    admin: Pubkey


def initialize_zeta_referrals_rewards_wallet(
    accounts: InitializeZetaReferralsRewardsWalletAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["referrals_rewards_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["usdc_mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf5\xe5\xdfx\x07\x86\xf7\xf8"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import logging
import re

import colorlog
from solana.utils.cluster import cluster_api_url

from zetamarkets_py import constants
from zetamarkets_py.types import Network


def convert_fixed_int_to_decimal(amount: int) -> float:
    return amount / 10**constants.PLATFORM_PRECISION


def convert_decimal_to_fixed_int(amount: float) -> int:
    return int((amount * 10**constants.PLATFORM_PRECISION / constants.TICK_SIZE)) * constants.TICK_SIZE


def convert_fixed_lot_to_decimal(amount: int) -> float:
    return amount / 10**constants.POSITION_PRECISION


def convert_decimal_to_fixed_lot(amount: float) -> int:
    return int(amount * 10**constants.POSITION_PRECISION)


def http_to_ws(endpoint: str) -> str:
    return re.sub(r"^http", "ws", endpoint)


def cluster_endpoint(network: Network, tls: bool = True, ws: bool = False) -> str:
    """Retrieve the RPC API URL for the specified cluster.

    :param cluster: The name of the cluster to use.
    :param tls: If True, use https. Defaults to True.
    """
    endpoint = cluster_api_url(network.value, tls=tls)  # type: ignore
    if ws:
        ws_endpoint = http_to_ws(endpoint)
        return ws_endpoint
    else:
        return endpoint


def get_tif_offset(expiry_ts: int, epoch_length: int, current_ts: int) -> int:
    if expiry_ts < current_ts:
        raise Exception(f"Cannot place expired order, current_ts: {current_ts}, expiry_ts: {expiry_ts}")
    epoch_start = current_ts - (current_ts % epoch_length)

    tif_offset = expiry_ts - epoch_start
    return min(tif_offset, epoch_length)


def create_logger(name: str, log_level: int = logging.CRITICAL, file_name: str = None) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    formatter = colorlog.ColoredFormatter("%(log_color)s%(levelname)s:%(name)s:%(message)s")

    # Create console handler and set level to same as the logger
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if file_name:
        fh = logging.FileHandler(file_name)
        fh.setLevel(log_level)
        file_formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s")
        fh.setFormatter(file_formatter)
        logger.addHandler(fh)
    return logger



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class InitializeMarketNodeArgsJSON(typing.TypedDict):
    nonce: int
    index: int


@dataclass
class InitializeMarketNodeArgs:
    layout: typing.ClassVar = borsh.CStruct("nonce" / borsh.U8, "index" / borsh.U8)
    nonce: int
    index: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeMarketNodeArgs":
        return cls(nonce=obj.nonce, index=obj.index)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"nonce": self.nonce, "index": self.index}

    def to_json(self) -> InitializeMarketNodeArgsJSON:
        return {"nonce": self.nonce, "index": self.index}

    @classmethod
    def from_json(cls, obj: InitializeMarketNodeArgsJSON) -> "InitializeMarketNodeArgs":
        return cls(nonce=obj["nonce"], index=obj["index"])



from .apply_perp_funding import (
    ApplyPerpFundingAccounts,
    ApplyPerpFundingArgs,
    apply_perp_funding,
)
from .cancel_all_market_orders import (
    CancelAllMarketOrdersAccounts,
    CancelAllMarketOrdersArgs,
    cancel_all_market_orders,
)
from .cancel_order import CancelOrderAccounts, CancelOrderArgs, cancel_order
from .cancel_order_by_client_order_id import (
    CancelOrderByClientOrderIdAccounts,
    CancelOrderByClientOrderIdArgs,
    cancel_order_by_client_order_id,
)
from .cancel_order_by_client_order_id_no_error import (
    CancelOrderByClientOrderIdNoErrorAccounts,
    CancelOrderByClientOrderIdNoErrorArgs,
    cancel_order_by_client_order_id_no_error,
)
from .cancel_order_halted import (
    CancelOrderHaltedAccounts,
    CancelOrderHaltedArgs,
    cancel_order_halted,
)
from .cancel_order_no_error import (
    CancelOrderNoErrorAccounts,
    CancelOrderNoErrorArgs,
    cancel_order_no_error,
)
from .claim_referrals_rewards import (
    ClaimReferralsRewardsAccounts,
    claim_referrals_rewards,
)
from .close_cross_margin_account import (
    CloseCrossMarginAccountAccounts,
    CloseCrossMarginAccountArgs,
    close_cross_margin_account,
)
from .close_cross_margin_account_manager import (
    CloseCrossMarginAccountManagerAccounts,
    close_cross_margin_account_manager,
)
from .close_open_orders_v3 import (
    CloseOpenOrdersV3Accounts,
    CloseOpenOrdersV3Args,
    close_open_orders_v3,
)
from .crank_event_queue import (
    CrankEventQueueAccounts,
    CrankEventQueueArgs,
    crank_event_queue,
)
from .deposit_v2 import DepositV2Accounts, DepositV2Args, deposit_v2
from .initialize_cross_margin_account import (
    InitializeCrossMarginAccountAccounts,
    InitializeCrossMarginAccountArgs,
    initialize_cross_margin_account,
)
from .initialize_cross_margin_account_manager import (
    InitializeCrossMarginAccountManagerAccounts,
    initialize_cross_margin_account_manager,
)
from .initialize_open_orders_v3 import (
    InitializeOpenOrdersV3Accounts,
    InitializeOpenOrdersV3Args,
    initialize_open_orders_v3,
)
from .initialize_referrer_account import (
    InitializeReferrerAccountAccounts,
    initialize_referrer_account,
)
from .initialize_referrer_alias import (
    InitializeReferrerAliasAccounts,
    InitializeReferrerAliasArgs,
    initialize_referrer_alias,
)
from .initialize_zeta_referrals_rewards_wallet import (
    InitializeZetaReferralsRewardsWalletAccounts,
    initialize_zeta_referrals_rewards_wallet,
)
from .liquidate_v2 import LiquidateV2Accounts, LiquidateV2Args, liquidate_v2
from .place_perp_order_v3 import (
    PlacePerpOrderV3Accounts,
    PlacePerpOrderV3Args,
    place_perp_order_v3,
)
from .refer_user import ReferUserAccounts, refer_user
from .withdraw_v2 import WithdrawV2Accounts, WithdrawV2Args, withdraw_v2



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class OverrideExpiryArgsJSON(typing.TypedDict):
    expiry_index: int
    active_ts: int
    expiry_ts: int


@dataclass
class OverrideExpiryArgs:
    layout: typing.ClassVar = borsh.CStruct("expiry_index" / borsh.U8, "active_ts" / borsh.U64, "expiry_ts" / borsh.U64)
    expiry_index: int
    active_ts: int
    expiry_ts: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "OverrideExpiryArgs":
        return cls(
            expiry_index=obj.expiry_index,
            active_ts=obj.active_ts,
            expiry_ts=obj.expiry_ts,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "expiry_index": self.expiry_index,
            "active_ts": self.active_ts,
            "expiry_ts": self.expiry_ts,
        }

    def to_json(self) -> OverrideExpiryArgsJSON:
        return {
            "expiry_index": self.expiry_index,
            "active_ts": self.active_ts,
            "expiry_ts": self.expiry_ts,
        }

    @classmethod
    def from_json(cls, obj: OverrideExpiryArgsJSON) -> "OverrideExpiryArgs":
        return cls(
            expiry_index=obj["expiry_index"],
            active_ts=obj["active_ts"],
            expiry_ts=obj["expiry_ts"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdateMarginParametersArgsJSON(typing.TypedDict):
    future_margin_initial: int
    future_margin_maintenance: int


@dataclass
class UpdateMarginParametersArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "future_margin_initial" / borsh.U64, "future_margin_maintenance" / borsh.U64
    )
    future_margin_initial: int
    future_margin_maintenance: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateMarginParametersArgs":
        return cls(
            future_margin_initial=obj.future_margin_initial,
            future_margin_maintenance=obj.future_margin_maintenance,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "future_margin_initial": self.future_margin_initial,
            "future_margin_maintenance": self.future_margin_maintenance,
        }

    def to_json(self) -> UpdateMarginParametersArgsJSON:
        return {
            "future_margin_initial": self.future_margin_initial,
            "future_margin_maintenance": self.future_margin_maintenance,
        }

    @classmethod
    def from_json(cls, obj: UpdateMarginParametersArgsJSON) -> "UpdateMarginParametersArgs":
        return cls(
            future_margin_initial=obj["future_margin_initial"],
            future_margin_maintenance=obj["future_margin_maintenance"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class ExpirySeriesJSON(typing.TypedDict):
    active_ts: int
    expiry_ts: int
    dirty: bool
    padding: list[int]


@dataclass
class ExpirySeries:
    layout: typing.ClassVar = borsh.CStruct(
        "active_ts" / borsh.U64,
        "expiry_ts" / borsh.U64,
        "dirty" / borsh.Bool,
        "padding" / borsh.U8[15],
    )
    active_ts: int
    expiry_ts: int
    dirty: bool
    padding: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "ExpirySeries":
        return cls(
            active_ts=obj.active_ts,
            expiry_ts=obj.expiry_ts,
            dirty=obj.dirty,
            padding=obj.padding,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "active_ts": self.active_ts,
            "expiry_ts": self.expiry_ts,
            "dirty": self.dirty,
            "padding": self.padding,
        }

    def to_json(self) -> ExpirySeriesJSON:
        return {
            "active_ts": self.active_ts,
            "expiry_ts": self.expiry_ts,
            "dirty": self.dirty,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: ExpirySeriesJSON) -> "ExpirySeries":
        return cls(
            active_ts=obj["active_ts"],
            expiry_ts=obj["expiry_ts"],
            dirty=obj["dirty"],
            padding=obj["padding"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class PositionJSON(typing.TypedDict):
    size: int
    cost_of_trades: int


@dataclass
class Position:
    layout: typing.ClassVar = borsh.CStruct("size" / borsh.I64, "cost_of_trades" / borsh.U64)
    size: int
    cost_of_trades: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "Position":
        return cls(size=obj.size, cost_of_trades=obj.cost_of_trades)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"size": self.size, "cost_of_trades": self.cost_of_trades}

    def to_json(self) -> PositionJSON:
        return {"size": self.size, "cost_of_trades": self.cost_of_trades}

    @classmethod
    def from_json(cls, obj: PositionJSON) -> "Position":
        return cls(size=obj["size"], cost_of_trades=obj["cost_of_trades"])



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdateZetaGroupExpiryArgsJSON(typing.TypedDict):
    expiry_interval_seconds: int
    new_expiry_threshold_seconds: int


@dataclass
class UpdateZetaGroupExpiryArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "expiry_interval_seconds" / borsh.U32,
        "new_expiry_threshold_seconds" / borsh.U32,
    )
    expiry_interval_seconds: int
    new_expiry_threshold_seconds: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateZetaGroupExpiryArgs":
        return cls(
            expiry_interval_seconds=obj.expiry_interval_seconds,
            new_expiry_threshold_seconds=obj.new_expiry_threshold_seconds,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "expiry_interval_seconds": self.expiry_interval_seconds,
            "new_expiry_threshold_seconds": self.new_expiry_threshold_seconds,
        }

    def to_json(self) -> UpdateZetaGroupExpiryArgsJSON:
        return {
            "expiry_interval_seconds": self.expiry_interval_seconds,
            "new_expiry_threshold_seconds": self.new_expiry_threshold_seconds,
        }

    @classmethod
    def from_json(cls, obj: UpdateZetaGroupExpiryArgsJSON) -> "UpdateZetaGroupExpiryArgs":
        return cls(
            expiry_interval_seconds=obj["expiry_interval_seconds"],
            new_expiry_threshold_seconds=obj["new_expiry_threshold_seconds"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class HaltArgsJSON(typing.TypedDict):
    spot_prices: list[int]
    timestamp: int


@dataclass
class HaltArgs:
    layout: typing.ClassVar = borsh.CStruct("spot_prices" / borsh.U64[5], "timestamp" / borsh.U64)
    spot_prices: list[int]
    timestamp: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "HaltArgs":
        return cls(spot_prices=obj.spot_prices, timestamp=obj.timestamp)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"spot_prices": self.spot_prices, "timestamp": self.timestamp}

    def to_json(self) -> HaltArgsJSON:
        return {"spot_prices": self.spot_prices, "timestamp": self.timestamp}

    @classmethod
    def from_json(cls, obj: HaltArgsJSON) -> "HaltArgs":
        return cls(spot_prices=obj["spot_prices"], timestamp=obj["timestamp"])



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class CloseCrossMarginAccountArgs(typing.TypedDict):
    subaccount_index: int


layout = borsh.CStruct("subaccount_index" / borsh.U8)


class CloseCrossMarginAccountAccounts(typing.TypedDict):
    cross_margin_account: Pubkey
    cross_margin_account_manager: Pubkey
    authority: Pubkey


def close_cross_margin_account(
    args: CloseCrossMarginAccountArgs,
    accounts: CloseCrossMarginAccountAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["cross_margin_account"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["cross_margin_account_manager"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcb\xc4\xbb<\r\xaa\xbeE"
    encoded_args = layout.build(
        {
            "subaccount_index": args["subaccount_index"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from construct import BitsSwapped, BitStruct, Container, Flag, Padding
from solders.pubkey import Pubkey

from zetamarkets_py.serum_client import types


@dataclass
class EventFlags:
    layout: typing.ClassVar = BitsSwapped(
        BitStruct(
            "fill" / Flag,
            "out" / Flag,
            "bid" / Flag,
            "maker" / Flag,
            Padding(4),
        )
    )
    fill: bool
    out: bool
    bid: bool
    maker: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "EventFlags":
        return cls(
            fill=obj.fill,
            out=obj.out,
            bid=obj.bid,
            maker=obj.maker,
        )


@dataclass
class Event:
    layout: typing.ClassVar = borsh.CStruct(
        "event_flags" / EventFlags.layout,
        "open_order_slot" / borsh.U8,
        "fee_tier" / borsh.U8,
        "padding" / borsh.U8[5],
        "native_quantity_released" / borsh.U64,
        "native_quantity_paid" / borsh.U64,
        "native_fee_or_rebate" / borsh.U64,
        "order_id" / borsh.U128,
        "public_key" / BorshPubkey,
        "client_order_id" / borsh.U64,
    )
    event_flags: EventFlags
    open_order_slot: int
    fee_tier: int
    native_quantity_released: int
    native_quantity_paid: int
    native_fee_or_rebate: int
    order_id: int
    public_key: Pubkey
    client_order_id: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "Event":
        return cls(
            event_flags=EventFlags.from_decoded(obj.event_flags),
            open_order_slot=obj.open_order_slot,
            fee_tier=obj.fee_tier,
            native_quantity_released=obj.native_quantity_released,
            native_quantity_paid=obj.native_quantity_paid,
            native_fee_or_rebate=obj.native_fee_or_rebate,
            order_id=obj.order_id,
            public_key=obj.public_key,
            client_order_id=obj.client_order_id,
        )


@dataclass
class QueueHeader:
    layout: typing.ClassVar = borsh.CStruct(
        "padding_1" / borsh.U8[5],
        "account_flags" / types.account_flags.AccountFlags.layout,
        "head" / borsh.U32,
        "padding_2" / borsh.U8[4],
        "count" / borsh.U32,
        "padding_3" / borsh.U8[4],
        "next_seq_num" / borsh.U32,
        "padding_4" / borsh.U8[4],
    )
    account_flags: types.account_flags.AccountFlags
    head: int
    count: int
    next_seq_num: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "QueueHeader":
        return cls(
            account_flags=types.account_flags.AccountFlags.from_decoded(obj.account_flags),
            head=obj.head,
            count=obj.count,
            next_seq_num=obj.next_seq_num,
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class NormalJSON(typing.TypedDict):
    kind: typing.Literal["Normal"]


class MarketMakerJSON(typing.TypedDict):
    kind: typing.Literal["MarketMaker"]


@dataclass
class Normal:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Normal"

    @classmethod
    def to_json(cls) -> NormalJSON:
        return NormalJSON(
            kind="Normal",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Normal": {},
        }


@dataclass
class MarketMaker:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "MarketMaker"

    @classmethod
    def to_json(cls) -> MarketMakerJSON:
        return MarketMakerJSON(
            kind="MarketMaker",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "MarketMaker": {},
        }


MarginAccountTypeKind = typing.Union[Normal, MarketMaker]
MarginAccountTypeJSON = typing.Union[NormalJSON, MarketMakerJSON]


def from_decoded(obj: dict) -> MarginAccountTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Normal" in obj:
        return Normal()
    if "MarketMaker" in obj:
        return MarketMaker()
    raise ValueError("Invalid enum object")


def from_json(obj: MarginAccountTypeJSON) -> MarginAccountTypeKind:
    if obj["kind"] == "Normal":
        return Normal()
    if obj["kind"] == "MarketMaker":
        return MarketMaker()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Normal" / borsh.CStruct(), "MarketMaker" / borsh.CStruct())
